# -*- coding: utf-8 -*-

"""Handling datasets"""

import os
import sys
import urllib.request

import numpy as np
import scipy.sparse as sp_sparse
from sklearn.preprocessing import StandardScaler
from neuralee._aux import ea, x2p
from tqdm import trange


class GeneExpressionDataset(object):
    """Gene Expression dataset. It deals with:
    - log_variational expression -> torch.log(1 + X)
    - local library size normalization (mean, var) per batch
    """

    def __init__(self, Y, batch_indices=None, labels=None, gene_names=None,
                 cell_types=None):

        if type(Y) is np.ndarray:
            self.Y = Y.astype(np.float32)
        else:
            self.Y = Y.toarray().astype(np.float32)
        self.nb_genes = self.Y.shape[1]
        if batch_indices is None:
            batch_indices = np.zeros((len(self), 1))
        if labels is None:
            labels = np.zeros((len(self), 1))
        self.batch_indices, self.n_batches = arrange_categories(batch_indices)
        self.labels, self.n_labels = arrange_categories(labels)

        if gene_names is not None:
            assert self.nb_genes == len(gene_names)
            self.gene_names = np.array(gene_names, dtype=np.str)
        if cell_types is not None:
            assert self.n_labels == len(cell_types)
            self.cell_types = np.array(cell_types, dtype=np.str)

    def __len__(self):
        return self.Y.shape[0]

    def log_shift(self):

        self.Y = np.log(self.Y + 1)

    def standardscale(self):
        self.std_scaler = StandardScaler()
        self.Y = self.std_scaler.fit_transform(self.Y)

    def affinity(self, aff='ea', perplexity=20.0):

        print('Compute affinity, perplexity={}, on entire dataset'.
              format(perplexity))
        self.Wp, self.Wn = self._affinity(self.Y, aff, perplexity)
        Lp = np.diagflat(self.Wp.sum(axis=1)) - self.Wp
        self.Lps = np.expand_dims(Lp, 0)
        self.Wns = np.expand_dims(self.Wn, 0)
        self.Y_splits = np.expand_dims(self.Y, 0)

    def affinity_split(self, N_small=None, aff='ea', perplexity=20.0,
                       verbose=False):

        N = len(self)
        if N_small is not None:
            N_small = N_small if isinstance(N_small, int) else int(N_small * N)
        print('Compute affinity, perplexity={}, N_small={}, on each batch'.
              format(perplexity, N_small))
        if N_small is None or N_small == N:
            return
        split = int(np.floor(N / N_small))
        indexes = np.random.permutation(N)[: split * N_small]
        self.Y_splits = self.Y[indexes].reshape(split, N_small, -1)
        self.Lps = np.zeros((split, N_small, N_small), dtype=np.float32)
        self.Wns = np.zeros((split, N_small, N_small), dtype=np.float32)
        with trange(split, desc="affinity on each batch", file=sys.stdout,
                    disable=not verbose) as pbar:
            for i in pbar:
                Wp, self.Wns[i] = \
                    self._affinity(self.Y_splits[i], aff, perplexity)
                self.Lps[i] = np.diagflat(Wp.sum(axis=1)) - Wp
                pbar.update(1)

    @staticmethod
    def _affinity(Y, aff, perplexity):
        """Return attractive and repulsive weights matrices.

        Parameters
        -----------
        Y : numpy.ndarray
            N x D matrix of gN groups of N row D-dimensional vectors.
        perplexity : scalar int
            the perplexity.s
        sparseK : torch.LongTensor, scalar
            the number of neighbors, None represents nonsparse.

        Returns:
        ----------
        Wp : torch.sparse.FloatTensor
            N x N attractive sparse weights matrix.
        Wn : torch.sparse.FloatTensor
             N x N repulsive sparse weights matrix.
        """

        N = Y.shape[0]
        assert aff in ['ea', 'x2p']
        if aff == 'ea':
            Wp, Wn = ea(Y, perplexity)
        else:
            Wp, Wn = x2p(Y, perplexity)

        Wp = (Wp + Wp.T) / (2 * N)
        Wn = (Wn + Wn.T) / (2 * Wn.sum(axis=None))

        return Wp, Wn

    def remove_zero_sample(self):
        ne_cells = self.Y.sum(axis=1) > 0
        to_keep = np.where(ne_cells)[0]
        if len(to_keep) < len(self):
            print("Cells with zero expression in all genes considered were "
                  "removed")
            self.update_cells(to_keep)

    def download_and_preprocess(self):
        self.download()
        return self.preprocess()

    def update_genes(self, subset_genes):
        new_n_genes = len(subset_genes) \
            if subset_genes.dtype is not np.dtype('bool') \
            else subset_genes.sum()
        print("Downsampling from %i to %i genes" %
              (self.nb_genes, new_n_genes))
        if hasattr(self, 'gene_names'):
            self.gene_names = self.gene_names[subset_genes]
        if hasattr(self, 'gene_symbols'):
            self.gene_symbols = self.gene_symbols[subset_genes]
        self.Y = self.Y[:, subset_genes]
        self.nb_genes = self.Y.shape[1]
        self.remove_zero_sample()

    def update_cells(self, subset_cells):
        new_n_cells = len(subset_cells) \
            if subset_cells.dtype is not np.dtype('bool') \
            else subset_cells.sum()
        print("Downsampling from %i to %i cells" % (len(self), new_n_cells))
        for attr_name in ['Y', 'labels', 'batch_indices']:
            setattr(self, attr_name, getattr(self, attr_name)[subset_cells])

    def subsample_genes(self, new_n_genes=None, subset_genes=None):
        n_cells, n_genes = self.Y.shape
        if subset_genes is None and \
           (new_n_genes is False or new_n_genes >= n_genes):
            # Do nothing if subsample more genes than total number of genes
            return None
        if subset_genes is None:
            std_scaler = StandardScaler(with_mean=False)
            std_scaler.fit(self.Y.astype(np.float64))
            subset_genes = np.argsort(std_scaler.var_)[::-1][:new_n_genes]
        self.update_genes(subset_genes)

    def filter_genes(self, gene_names_ref, on='gene_names'):
        """
        Same as _filter_genes but overwrites on current dataset instead of
        returning data, and updates genes names and symbols
        """
        subset_genes = GeneExpressionDataset.__filter_genes(
            self, gene_names_ref, on=on)
        self.update_genes(subset_genes)

    def subsample_cells(self, size=1.):
        n_cells, n_genes = self.Y.shape
        new_n_cells = int(size * n_genes) if type(size) is not int else size
        indices = np.argsort(
            np.array(self.Y.sum(axis=1)).ravel())[::-1][:new_n_cells]
        self.update_cells(indices)

    def _cell_type_idx(self, cell_types):
        if type(cell_types[0]) is not int:
            cell_types_idx = \
                [np.where(cell_type == self.cell_types)[0][0]
                 for cell_type in cell_types]
        else:
            cell_types_idx = cell_types
        return np.array(cell_types_idx, dtype=np.int64)

    def _gene_idx(self, genes):
        if type(genes[0]) is not int:
            genes_idx = \
                [np.where(gene == self.gene_names)[0][0] for gene in genes]
        else:
            genes_idx = genes
        return np.array(genes_idx, dtype=np.int64)

    def filter_cell_types(self, cell_types):
        """
        :param cell_types:
            numpy array of type np.int (indices) ornp.str (cell-types names)
        :return:
        """
        cell_types_idx = self._cell_type_idx(cell_types)
        if hasattr(self, 'cell_types'):
            self.cell_types = self.cell_types[cell_types_idx]
            print("Only keeping cell types: \n" +
                  '\n'.join(list(self.cell_types)))
        idx_to_keep = []
        for idx in cell_types_idx:
            idx_to_keep += [np.where(self.labels == idx)[0]]
        self.update_cells(np.concatenate(idx_to_keep))
        self.labels, self.n_labels = \
            arrange_categories(self.labels, mapping_from=cell_types_idx)

    def merge_cell_types(self, cell_types, new_cell_type_name):
        """
        Merge some cell types into a new one, a change the labels accordingly.
        :param merge_cell_types:
            numpy array of type np.int (indices) or np.str (cell-types names)
        :return:
        """
        cell_types_idx = self._cell_type_idx(cell_types)
        for idx_from in zip(cell_types_idx):
            # Put at the end the new merged cell-type
            self.labels[self.labels == idx_from] = len(self.labels)
        self.labels, self.n_labels = arrange_categories(self.labels)
        if hasattr(self, 'cell_types') and type(cell_types[0]) is not int:
            new_cell_types = list(self.cell_types)
            for cell_type in cell_types:
                new_cell_types.remove(cell_type)
            new_cell_types.append(new_cell_type_name)
            self.cell_types = np.array(new_cell_types)

    def map_cell_types(self, cell_types_dict):
        """
        A map for the cell types to keep, and optionally merge together under
        a new name (value in the dict)
        :param cell_types_dict:
            a dictionary with tuples (str or int) as input
            and value (str or int) as output
        """
        keys = [(key,) if type(key) is not tuple else key
                for key in cell_types_dict.keys()]
        cell_types = [cell_type for cell_types in keys
                      for cell_type in cell_types]
        self.filter_cell_types(cell_types)
        for cell_types, new_cell_type_name in cell_types_dict.items():
            self.merge_cell_types(cell_types, new_cell_type_name)

    def download(self):
        if hasattr(self, 'urls') and hasattr(self, 'download_names'):
            for url, download_name in zip(self.urls, self.download_names):
                GeneExpressionDataset._download(
                    url, self.save_path, download_name)
        elif hasattr(self, 'url') and hasattr(self, 'download_name'):
            GeneExpressionDataset._download(
                self.url, self.save_path, self.download_name)

    @staticmethod
    def _download(url, save_path, download_name):
        if os.path.exists(os.path.join(save_path, download_name)):
            print("File %s already downloaded" %
                  (os.path.join(save_path, download_name)))
            return
        if url is None:
            print("You are trying to load a local file named %s and located at"
                  " %s but this file was not found at the location %s" %
                  (download_name, save_path,
                   os.path.join(save_path, download_name)))
        r = urllib.request.urlopen(url)
        print("Downloading file at %s" %
              os.path.join(save_path, download_name))

        def readIter(f, blocksize=1000):
            """Given a file 'f', returns an iterator that returns bytes of
            size 'blocksize' from the file, using read()."""
            while True:
                data = f.read(blocksize)
                if not data:
                    break
                yield data

        # Create the path to save the data
        if not os.path.exists(save_path):
            os.makedirs(save_path)

        with open(os.path.join(save_path, download_name), 'wb') as f:
            for data in readIter(r):
                f.write(data)

    @staticmethod
    def get_attributes_from_matrix(X, batch_indices=0, labels=None):
        ne_cells = X.sum(axis=1) > 0
        to_keep = np.where(ne_cells)
        if not ne_cells.all():
            X = X[to_keep]
            removed_idx = np.where(~ne_cells)[0]
            print("Cells with zero expression in all genes considered were "
                  "removed, the indices of the removed cells in the expression"
                  " matrix were:")
            print(removed_idx)
        batch_indices = batch_indices * np.ones((X.shape[0], 1)) \
            if type(batch_indices) is int else batch_indices[to_keep]
        labels = labels[to_keep].reshape(-1, 1) \
            if labels is not None else np.zeros_like(batch_indices)
        return X, batch_indices, labels

    @staticmethod
    def get_attributes_from_list(Xs, list_batches=None, list_labels=None):
        nb_genes = Xs[0].shape[1]
        assert all(X.shape[1] == nb_genes for X in Xs), \
            "All tensors must have same size"

        new_Xs = []
        batch_indices = []
        labels = []
        for i, X in enumerate(Xs):
            ne_cells = X.sum(axis=1) > 0
            to_keep = np.where(ne_cells)
            if len(to_keep) < X.shape[0]:
                removed_idx = np.where(~ne_cells)[0]
                print("Cells with zero expression in all genes considered were"
                      " removed, the indices of the removed cells in the "
                      "expression matrix were:")
                print(removed_idx)
            X = X[to_keep]
            new_Xs += [X]
            batch_indices += \
                [list_batches[i][to_keep] if list_batches is not None
                 else i * np.ones((X.shape[0], 1))]
            labels += \
                [list_labels[i][to_keep] if list_labels is not None
                 else np.zeros((X.shape[0], 1))]

        X = np.concatenate(new_Xs) if type(new_Xs[0]) is np.ndarray \
            else sp_sparse.vstack(new_Xs)
        batch_indices = np.concatenate(batch_indices)
        labels = np.concatenate(labels)
        return X, batch_indices, labels

    @staticmethod
    def concat_datasets(*gene_datasets, on='gene_names', shared_labels=True,
                        shared_batches=False):
        """
        Combines multiple unlabelled gene_datasets based on the intersection of
        gene names intersection. Datasets should all have
        gene_dataset.n_labels=0. Batch indices are generated in the same order
        as datasets are given.
        :param gene_datasets: a sequence of gene_datasets object
        :return: a GeneExpressionDataset instance of the concatenated datasets
        """
        assert all([hasattr(gene_dataset, on)
                    for gene_dataset in gene_datasets])

        gene_names_ref = \
            set.intersection(*[set(getattr(gene_dataset, on))
                               for gene_dataset in gene_datasets])
        # keep gene order of the first dataset
        gene_names_ref = \
            [gene_name for gene_name in getattr(gene_datasets[0], on)
             if gene_name in gene_names_ref]
        print("Keeping %d genes" % len(gene_names_ref))

        Xs = [GeneExpressionDataset._filter_genes(
            dataset, gene_names_ref, on=on)[0] for dataset in gene_datasets]
        if gene_datasets[0].dense:
            X = np.concatenate(
                [X if type(X) is np.ndarray else X.A for X in Xs])
        else:
            X = sp_sparse.vstack(
                [X if type(X) is not np.ndarray else sp_sparse.csr_matrix(X)
                 for X in Xs])

        batch_indices = np.zeros((X.shape[0], 1))
        n_batch_offset = 0
        current_index = 0
        for gene_dataset in gene_datasets:
            next_index = current_index + len(gene_dataset)
            batch_indices[current_index:next_index] = \
                gene_dataset.batch_indices + n_batch_offset
            n_batch_offset += \
                (gene_dataset.n_batches if not shared_batches else 0)
            current_index = next_index

        cell_types = None
        if shared_labels:
            if all([hasattr(gene_dataset, "cell_types")
                    for gene_dataset in gene_datasets]):
                cell_types = list(
                    set([cell_type for gene_dataset in gene_datasets
                        for cell_type in gene_dataset.cell_types])
                )
                labels = []
                for gene_dataset in gene_datasets:
                    mapping = [cell_types.index(cell_type)
                               for cell_type in gene_dataset.cell_types]
                    labels += [arrange_categories(
                        gene_dataset.labels, mapping_to=mapping)[0]]
                labels = np.concatenate(labels)
            else:
                labels = \
                    np.concatenate([gene_dataset.labels
                                    for gene_dataset in gene_datasets])
        else:
            labels = np.zeros((X.shape[0], 1))
            n_labels_offset = 0
            current_index = 0
            for gene_dataset in gene_datasets:
                next_index = current_index + len(gene_dataset)
                labels[current_index:next_index] = \
                    gene_dataset.labels + n_labels_offset
                n_labels_offset += gene_dataset.n_labels
                current_index = next_index

        result = GeneExpressionDataset(
            X, batch_indices, labels, gene_names=gene_names_ref,
            cell_types=cell_types)
        result.barcodes = [gene_dataset.barcodes
                           if hasattr(gene_dataset, 'barcodes') else None
                           for gene_dataset in gene_datasets]
        return result

    @staticmethod
    def _filter_genes(gene_dataset, gene_names_ref, on='gene_names'):
        """
        :return:
            gene_dataset.X filtered by the
            corresponding genes ( / columns / features), idx_genes
        """
        gene_names = list(getattr(gene_dataset, on))
        subset_genes = np.array([gene_names.index(gene_name)
                                 for gene_name in gene_names_ref],
                                dtype=np.int64)
        return gene_dataset.Y[:, subset_genes], subset_genes

    @staticmethod
    def __filter_genes(gene_dataset, gene_names_ref, on='gene_names'):
        """
        :return:
            gene_dataset.X filtered by the
            corresponding genes ( / columns / features), idx_genes
        """
        gene_names = list(getattr(gene_dataset, on))
        subset_genes = np.array([gene_names.index(gene_name)
                                 for gene_name in gene_names_ref],
                                dtype=np.int64)
        return subset_genes


def arrange_categories(original_categories, mapping_from=None,
                       mapping_to=None):
    unique_categories = np.unique(original_categories)
    n_categories = len(unique_categories)
    if mapping_to is None:
        mapping_to = range(n_categories)
    if mapping_from is None:
        mapping_from = unique_categories
    # one cell_type can have no instance in dataset
    assert n_categories <= len(mapping_from)
    assert len(mapping_to) == len(mapping_from)

    new_categories = np.copy(original_categories)
    for idx_from, idx_to in zip(mapping_from, mapping_to):
        new_categories[original_categories == idx_from] = idx_to
    return new_categories.astype(int), n_categories



import time
import copy
import sys
import os
from tqdm import trange

import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np

from ._aux import error_ee, error_ee_cpu, error_ee_cuda, ls_ee, eloss
from torch.utils.data import DataLoader, TensorDataset


class FCLayers(nn.Module):
    def __init__(self, di, do):
        super(FCLayers, self).__init__()

        self.net = nn.Sequential(

            nn.Linear(di, 50),
            nn.BatchNorm1d(50),
            nn.ReLU(),
            nn.Linear(50, 50),
            nn.BatchNorm1d(50),
            nn.ReLU(),
            nn.Linear(50, do),
        )
        for m in self.net.modules():
            if isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, mean=0, std=0.1)
                nn.init.constant_(m.bias, val=0)

    def forward(self, y):
        return self.net(y)


class NeuralEE(object):
    """NeuralEE method.

    Attributes:
    ----------
    dataset : GeneExpressionDataset
        main attribute Y of dataset is N x D training matrix.
        N is dataset size, D is high original feature dimension.
    d : scalar, int, optional
        low embedded dimension. Default: 2.
    lam : scalar, optional
        lambda value for elastic embedding function. Default: 10.
    device : torch.device, optional
        device chosen to compute. Default: torch.device('cpu').

    """
    def __init__(self, dataset, d=2, lam=10, device=None):
        self.dataset = dataset
        self.d = d
        self.lam = lam
        self.device = device if device is not None else torch.device('cpu')
        self.Y = torch.from_numpy(self.dataset.Y)
        self.Y_splits = torch.from_numpy(self.dataset.Y_splits) \
            if hasattr(self.dataset, 'Y_splits') else None
        self.Wp = torch.from_numpy(self.dataset.Wp) \
            if hasattr(self.dataset, 'Wp') else None
        self.Wn = torch.from_numpy(self.dataset.Wn) \
            if hasattr(self.dataset, 'Wn') else None
        self.Lps = torch.from_numpy(self.dataset.Lps) \
            if hasattr(self.dataset, 'Lps') else None
        self.Wns = torch.from_numpy(self.dataset.Wns) \
            if hasattr(self.dataset, 'Wns') else None

    def __len__(self):
        return len(self.dataset)

    @property
    def D(self):
        return self.Y.shape[1]

    @property
    def labels(self):
        return self.dataset.labels.squeeze()

    @staticmethod
    def _loss(Y, Wp, Wn, lam, net, device, calculate_error):
        results = dict()
        with torch.no_grad():
            net.eval()
            net.to(device)
            X = net(Y.to(device))
        results['X'] = X.cpu()
        if calculate_error is not None:
            assert calculate_error in ['cpu', 'cuda']
            if calculate_error == 'cpu':
                e = error_ee_cpu(X.cpu().numpy(), Wp, Wn, lam)
                results['e'] = e
            else:
                e = error_ee_cuda(X, Wp, Wn, lam)
                results['e'] = e.item()

        return results

    def _loss_entire(self, calculate_error=None):
        return self._loss(
            self.Y, self.Wp, self.Wn, self.lam, self.net, self.device,
            calculate_error=calculate_error)

    def EE(self, size=1., maxit=200, tol=1e-5, frequence=None, aff='ea',
           perplexity=20.0):
        """Free Elastic embedding (no mapping).

        Fast training of nonlinear embeddings using the spectral direction for
        the Elastic Embedding (EE) algorithm.

        Reference:
            Partial-Hessian Strategies for Fast Learning of Nonlinear
            Embeddings.
            http://faculty.ucmerced.edu/mcarreira-perpinan/papers/icml12.pdf

        Parameters
        -----------
        size : scalar, int or float in [0, 1], optional
            subsample of entire dataset to embed. Default: 1.0, not subsample.
            If subsample, the affinity will be recalculated on subsmaples.
        maxit : scalar, int, optional
            max number of iterations for EE. Default: 200.
        tol : scalar, optional
            minimum relative distance between consecutive X. Default: 1e-5.
        frequence : None or int, scalar, optional
            frequence to display iterating results. Defualt: None, not display.
        aff : {'ea', 'x2p'}, optional
            affinity used to calculate attractive weights if subsampled.
            Default: 'ea'.
        perplexity : scalar, optional
            perplexity defined in elastic embedding function if subsampled.
            Default: 20.0.

        Returns
        ----------
        results : dict
            X : embedding results.
            e : embedding loss.
            sub_samples : subsamples information if subsampled.
                Y : original feature on subsamples.
                labels : labels on subsamples.
                Wp : attractive weights on subsamples.
                Wn : repulsive weights on subsamples.
        """

        since = time.time()
        N = len(self)
        results = dict()
        N_sub = size if isinstance(size, int) else int(N * size)
        assert N_sub <= 15000, \
            'Number of samples is too huge for free EE.'
        X0 = 1e-5 * torch.randn(N_sub, self.d)
        if N_sub == N:
            assert self.Wp is not None, \
                'affinity on entire dataset is needed.'
            Wp = self.Wp
            Wn = self.Wn
        else:
            ind_sub = torch.randperm(N)[: N_sub].tolist()
            Y = self.dataset.Y[ind_sub]
            labels = self.lamabels[ind_sub].squeeze()
            print('Compute affinity on downsample')
            Wp, Wn = self.dataset._affinity(Y, aff, perplexity)
            Wp = torch.from_numpy(Wp)
            Wn = torch.from_numpy(Wn)
            sub_samples = {'Y': torch.from_numpy(Y),
                           'labels': labels,
                           'Wp': Wp,
                           'Wn': Wn}
            results['sub_samples'] = sub_samples

        Wp = Wp.to(self.device)
        Wn = Wn.to(self.device)
        Dp = Wp.sum(dim=1).diagflat()
        Lp4 = 4 * (-Wp + Dp)
        R = torch.cholesky(
            Lp4 + 1e-6 * torch.eye(N_sub, N_sub, device=self.device),
            upper=True)
        invR = R.inverse()
        S = torch.eye(N_sub, N_sub, device=self.device)
        P0 = -S @ invR @ invR.t() @ S.t()
        del R, S, invR, Dp
        torch.cuda.empty_cache()
        Xold = X0.to(self.device)
        e, ker = error_ee(Xold, Wp, Wn, self.lam)
        j = 1
        a = 1
        convcrit = maxit >= 1
        while convcrit:
            WWn = self.lam * Wn * ker
            DDn = WWn.sum(dim=1).diagflat()
            # gradient
            G = (Lp4 - 4 * (-WWn + DDn)) @ Xold
            P = P0 @ G  # spectral direction
            # line search
            X, e, ker, a = ls_ee(Xold, Wp, Wn, self.lam, P, e, G, a)

            convcrit = (j < maxit) and \
                (torch.norm(X - Xold) > tol * torch.norm(Xold))

            Xold = X
            if frequence is not None and j % frequence == 0:
                print('Epoch {}, EE loss is {:.6f}'.format(j, e))
            j += 1
        print('Elastic Embedding, lambda={}, completed in {:.2f}s, '
              'EE loss is {:.6f}'.format(
                self.lam, time.time() - since, e.item()))

        results['X'] = X.cpu()
        results['e'] = e.item()
        torch.cuda.empty_cache()
        return results

    def collate_fn(self, batch):
        y, Lp, Wn = batch[0]
        return y.to(self.device), Lp.to(self.device), Wn.to(self.device)

    def fine_tune(self, optimizer=None, size=1., net=None, frequence=50,
                  verbose=False, maxit=500, calculate_error='cuda',
                  pin_memory=True, aff='ea', perplexity=20.0,
                  save_embedding=None):
        """

        Parameters
        -----------
        optimizer : torch.optim, optional
            optimization for traning neural networks.
            Default: torch.optim.Adam(lr=0.01).
        size : scalar, int or float in [0, 1], optional
            subsample of entire dataset to embed.
            Default: 1.0, not subsample.
            If subsample, the affinity will be recalculated on subsmaples.
        net : torch.nn.Module, optional
            neural networks module as embedding function. Default: FCLayers.
        frequence : None or int, scalar, optional
            frequence to compare and save iterating results. Defualt: 50.
        verbose : bool, optional
            whether to show verbose training loss. Default: False.
        maxit : scalar, int, optional
            max number of iterations for fine tune. Default: 500.
        calculate_error:  {None, 'cuda', 'cpu'}
            device to compute error.
            if None, not compute and frequence ignored. Default: 'cuda'
        pin_memory : bool, optional
            whether to pin data on GPU memory to save time of transfer,
            which depends on your GPU memory. Default: True.
        aff : {'ea', 'x2p'}, optional
            affinity used to calculate attractive weights if subsampled.
            Default: 'ea'.
        perplexity : scalar, optional
            perplexity defined in elastic embedding function if subsampled.
            Default: 20.0.
        save_embedding: bool, optional
            whether to save iterating results accoring to frequence.
            Default: False.

        Returns
        ----------
        results : dict
            X : embedding results.
            e : embedding loss.
            sub_samples : subsamples information if subsampled.
                Y : original feature on subsamples.
                labels : labels on subsamples.
                Wp : attractive weights on subsamples.
                Wn : repulsive weights on subsamples.
        """
        since = time.time()
        results = dict()
        N = len(self)
        N_sub = size if isinstance(size, int) else int(N * size)

        if N_sub != N:
            print('Compute affinity on downsample')
            ind_sub = torch.randperm(N)[: N_sub].tolist()
            Y = self.dataset.Y[ind_sub]
            Wp, Wn = self.dataset._affinity(Y, aff, perplexity)
            Y, = torch.from_numpy(Y)
            Wp = torch.from_numpy(Wp)
            Wn = torch.from_numpy(Wn)
            labels = self.lamabels[ind_sub].squeeze()
            sub_samples = {'Y': Y,
                           'labels': labels,
                           'Wp': Wp,
                           'Wn': Wn}
            results['sub_samples'] = sub_samples
            Y_splits = Y.unsqueeze(0).to(self.device)
            Wp = Wp.to(self.device)
            Lps = (Wp.sum(dim=1).diagflat() - Wp).unsqueeze(0)
            Wns = Wn.unsqueeze(0).to(self.device)
        else:
            assert calculate_error is None or self.Wp is not None, \
                'affinity on entire dataset is needed.'
            assert self.Lps is not None, \
                'affinity is needed.'
            Lps = self.Lps.to(self.device) if pin_memory else self.Lps
            Wns = self.Wns.to(self.device) if pin_memory else self.Wns
            Y_splits = self.Y_splits.to(self.device) \
                if pin_memory else self.Y_splits

        dataset = TensorDataset(Y_splits, Lps, Wns)
        # for speed, only batch_size = 1
        if N_sub != N or pin_memory:
            dataloader = DataLoader(dataset, collate_fn=lambda batch: batch[0])
        else:
            dataloader = DataLoader(dataset, collate_fn=self.collate_fn)

        if save_embedding is not None:
            path = save_embedding + self.dataset.__class__.__name__ + '/'
            if not os.path.exists(path):
                os.makedirs(path)

        flag_pbar = calculate_error is None and verbose

        if net is not None:
            self.net = net
        elif hasattr(self, 'net'):
            pass
        else:
            self.net = FCLayers(self.D, self.d)

        if optimizer is None:
            optimizer = optim.Adam(self.net.parameters(), lr=0.01)

        if calculate_error is not None and frequence is not None:
            best_model_wts = copy.deepcopy(self.net.state_dict())
            best_loss = float('inf')

        self.net.train()
        self.net.to(self.device)
        with trange(maxit, desc="NeuralEE", file=sys.stdout,
                    disable=not flag_pbar) as pbar:
            for epoch in range(1, maxit + 1):
                for inputs, Lp_batch, Wn_batch in dataloader:
                    optimizer.zero_grad()
                    outputs = self.net(inputs)
                    loss = eloss(outputs, Lp_batch, Wn_batch, self.lam)
                    loss.backward()
                    optimizer.step()
                if frequence is not None and epoch % frequence == 0:
                    if calculate_error is not None:
                        if N_sub == N:
                            e = self._loss_entire(
                                calculate_error=calculate_error)['e']
                        else:
                            e = self._loss(
                                Y_splits[0], Wp, Wns[0], self.lam, self.net,
                                self.device, calculate_error)['e']
                        if verbose:
                            print(
                                'Epoch {}, EE loss is {:.6f}'.format(epoch, e))
                        if e < best_loss:
                            best_loss = e
                            best_model_wts = copy.deepcopy(
                                self.net.state_dict())
                    if save_embedding is not None:
                        X = self._loss_entire(
                            calculate_error=None)['X'].numpy()
                        np.save(
                            path + self.dataset.__class__.__name__ +
                            str(epoch), X)
                pbar.update(1)

        if calculate_error is not None and frequence is not None:
            self.net.load_state_dict(best_model_wts)
        if N_sub == N:
            results.update(self._loss_entire(calculate_error=calculate_error))
        else:
            results.update(self._loss(Y_splits[0], Wp, Wns[0], self.lam,
                           self.net, self.device, calculate_error))
        if calculate_error is not None:
            print('Neural Elastic Embedding, lambda={}, completed in {:.2f}s, '
                  'EE loss is {:.6f}'.format(
                    self.lam, time.time() - since, results['e']))
        else:
            print('Neural Elastic Embedding, lambda={}, completed in {:.2f}s.'
                  .format(self.lam, time.time() - since))
        torch.cuda.empty_cache()
        return results

    def map(self, samples=dict(), calculate_error='cuda'):
        """

        Parameters
        -----------
        samples : dict, optional
            Y : samples to map into low-dimensional coordinate.
            labels : samples labels. None is acceptable.
            Wp : attractive weights on samples.
                 None if error need not be calculated.
            Wn : repulsive weights on samples.
                 None if error need not be calculated.
            Default: dict(), mapping on training data.

        calculate_error:  {None, 'cuda', 'cpu'}
            device to compute error.
            if None, not compute and frequence ignored. Default: 'cuda'

        Returns
        ----------
        results : dict
            X : embedding results.
            e : embedding loss.
        """

        if samples == dict():
            assert calculate_error is None or hasattr(self.dataset, 'Wp'), \
                'affinity on entire dataset is needed.'
            samples = {'Y': torch.from_numpy(self.dataset.Y),
                       'labels': self.lamabels,
                       'Wp': None
                       if calculate_error is None else self.dataset.Wp,
                       'Wn': None
                       if calculate_error is None else self.dataset.Wn}
        results = self._loss(
            samples['Y'], samples['Wp'], samples['Wn'], self.lam, self.net,
            self.device, calculate_error)
        if calculate_error is not None:
            print('EE loss is {:.6f}'.format(results['e']))
        torch.cuda.empty_cache()
        return results



from .error import error_ee


def ls_ee(X, Wp, Wn, lam, P, ff, G, alpha0=1, rho=0.8, c=1e-1):
    """Backtracking line search for EE

    Reference: procedure 3.1, p. 41ff in:
        Nocedal and Wright: "Numerical Optimization", Springer, 1999.
    """

    alpha = alpha0
    tmp = c * G.view(-1).dot(P.view(-1))
    e, ker = error_ee(X + alpha * P, Wp, Wn, lam)
    while e > ff + alpha * tmp:
        alpha = rho * alpha
        e, ker = error_ee(X + alpha * P, Wp, Wn, lam)

    Xnew = X + alpha * P

    return Xnew, e, ker, alpha



from .affinity import ea, x2p
from .error import error_ee, ker_ee, error_ee_from_ker, error_ee_cpu, \
                   error_ee_cuda
from .gradient import gradient_ee
from .linearsearch import ls_ee
from .visualize import scatter, scatter_with_colorbar, scatter_without_outlier
from .embeddingsLoss import eloss

__all__ = ['ea',
           'x2p',
           'error_ee',
           'ker_ee',
           'error_ee_from_ker',
           'error_ee_cpu',
           'error_ee_cuda',
           'gradient_ee',
           'ls_ee',
           'scatter',
           'scatter_without_outlier',
           'scatter_with_colorbar',
           'eloss']



import os
from pathlib import Path
from zipfile import ZipFile

import numpy as np
import pandas as pd

from .dataset import GeneExpressionDataset


class HematoDataset(GeneExpressionDataset):
    r""" Loads hemato dataset.

    This dataset with continuous gene expression variations from hematopoeitic
    progenitor cells [31] contains 4,016 cells and 7,397 genes. We removed the
    library basal-bm1 which was of poor quality based on authors
    recommendation. We use their population balance analysis result as a
    potential function for differentiation.

    Args:
        :save_path: Save path of raw data file. Default: ``'data/'``.

    Examples:
        >>> gene_dataset = HematoDataset()

    """

    def __init__(self, save_path='data/HEMATO/'):
        self.save_path = save_path

        # file and datazip file
        self.urls = ["https://www.ncbi.nlm.nih.gov/geo/download/"
                     "?acc=GSM2388072&format=file&file="
                     "GSM2388072%5Fbasal%5Fbone%5Fmarrow%2Eraw%5Fumifm"
                     "%5Fcounts%2Ecsv%2Egz",
                     "https://github.com/romain-lopez/scVI-reproducibility/"
                     "raw/master/additional/data.zip"]

        # Originally: GSM2388072_basal_bone_marrow.raw_umifm_counts.csv.gz
        self.download_names = ["bBM.raw_umifm_counts.csv.gz", "data.zip"]
        self.gene_names_filename = "bBM.filtered_gene_list.paper.txt"
        self.spring_and_pba_filename = "bBM.spring_and_pba.csv"

        expression_data, gene_names, labels, self.y_spring, self.y_spring = \
            self.download_and_preprocess()

        super().__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                expression_data, labels=labels), gene_names=gene_names,

        )

        self.cell_types_levels = \
            ['Erythroid', 'Granulocytic Neutrophil', 'Lymphocytic',
             'Dendritic', 'Megakaryocytic', 'Monocytic', 'Basophilic']

    def preprocess(self):
        print("Preprocessing Hemato data")

        if len(os.listdir(self.save_path)) == 2:  # nothing extracted yet
            with ZipFile(os.path.join(self.save_path, 'data.zip'), 'r') as zip:
                zip.extractall(path=Path(self.save_path).parent)
        raw_counts = pd.read_csv(
            os.path.join(self.save_path, self.download_names[0]),
            compression='gzip')

        # remove this library to avoid dealing with batch effects
        raw_counts.drop(
            raw_counts.index[raw_counts["library_id"] == "basal_bm1"],
            inplace=True)

        spring_and_pba = pd.read_csv(
            os.path.join(self.save_path, self.spring_and_pba_filename))
        # with open(self.save_path + self.gene_names_filename) as f:
        #     gene_filter_list = f.read()
        #
        # gene_names = gene_filter_list.splitlines()
        gene_names = np.loadtxt(
            os.path.join(self.save_path, self.gene_names_filename),
            dtype=np.str)

        data = raw_counts.merge(spring_and_pba, how="inner")
        expression_data = data[gene_names]
        x_spring = data["x_spring"].values
        y_spring = data["y_spring"].values

        self.meta = data[["Potential", "Pr_Er", "Pr_Gr", "Pr_Ly", "Pr_DC",
                          "Pr_Mk", "Pr_Mo", "Pr_Ba"]]

        def logit(p):
            p = np.copy(p.values)
            p[p == 0] = np.min(p[p > 0])
            p[p == 1] = np.max(p[p < 1])
            return np.log(p / (1 - p))

        labels = logit(self.meta.iloc[:, 2]) - logit(self.meta.iloc[:, 1])
        expression_data = expression_data.values

        print("Finished preprocessing Hemato data")
        return expression_data, gene_names, labels, x_spring, y_spring



# On the 10X website:
# The main categories (Cell Ranger 1.1.0 / Cell Ranger 2.1.0 / ...)
# have same access suffix for each of their dataset.
# For dataset name (eg. 'pbmc8k', 'pbmc4k', ect...) their are two available
# specifications, either filtered or raw data
import os
import pickle
import tarfile

import numpy as np
import pandas as pd
from scipy import io
from scipy.sparse import csr_matrix

from .dataset import GeneExpressionDataset

available_datasets = {"1.1.0":
                      ["frozen_pbmc_donor_a",
                       "frozen_pbmc_donor_b",
                       "frozen_pbmc_donor_c",
                       "fresh_68k_pbmc_donor_a",
                       "cd14_monocytes",
                       "b_cells",
                       "cd34",
                       "cd56_nk",
                       "cd4_t_helper",
                       "regulatory_t",
                       "naive_t",
                       "memory_t",
                       "cytotoxic_t",
                       "naive_cytotoxic"
                       ],
                      "2.1.0":
                      ["pbmc8k",
                       "pbmc4k",
                       "t_3k",
                       "t_4k",
                       "neuron_9k"]}

to_groups = dict([(dataset_name, group)
                  for group, list_datasets in available_datasets.items()
                  for dataset_name in list_datasets])
available_specification = ['filtered', 'raw']


class Dataset10X(GeneExpressionDataset):
    r""" Loads a file from `10x`_ website.

    Args:
        :filename: Name of the dataset file.
        :save_path: Save path of the dataset. Default: ``'data/'``.
        :type: Either `filtered` data or `raw` data. Default: ``'filtered'``.
        :subset_genes: List of genes for subsampling. Default: ``None``.
        :dense: Whether to load as dense or sparse. Default: ``False``.
        :remote: Whether the 10X dataset is to be downloaded from the website
                 or whether it is a local dataset, if
        remote is False then os.path.join(save_path, filename) must be the path
        to the directory that contains matrix.mtx and genes.tsv files

    Examples:
        >>> tenX_dataset = Dataset10X("neuron_9k")

    .. _10x:
        http://cf.10xgenomics.com/

    """

    def __init__(self, filename, save_path='data/', type='filtered',
                 dense=False, remote=True, genecol=0):

        self.remote = remote
        self.save_path = save_path
        self.genecol = genecol
        if self.remote:
            group = to_groups[filename]
            self.url = ("http://cf.10xgenomics.com/samples/cell-exp"
                        "/%s/%s/%s_%s_gene_bc_matrices.tar.gz" %
                        (group, filename, filename, type))
            self.save_path = os.path.join(save_path, '10X/%s/' % filename)
            self.save_name = '%s_gene_bc_matrices' % type
            self.download_name = self.save_name + '.tar.gz'
        else:
            try:
                assert os.path.isdir(os.path.join(self.save_path, filename))
            except AssertionError:
                print("The file %s was not found in the location you gave" %
                      filename)
                raise
            self.save_path = os.path.join(self.save_path, filename)

        self.dense = dense

        expression_data, gene_names = self.download_and_preprocess()
        super().__init__(*GeneExpressionDataset.get_attributes_from_matrix(
            expression_data), gene_names=gene_names)

    def preprocess(self):
        print("Preprocessing dataset")
        path = self.save_path
        if self.remote:
            if len(os.listdir(self.save_path)) == 1:  # nothing extracted yet
                print("Extracting tar file")
                tar = tarfile.open(
                  os.path.join(self.save_path, self.download_name), "r:gz")
                tar.extractall(path=self.save_path)
                tar.close()

            path = (os.path.join(self.save_path,
                    [name for name in os.listdir(self.save_path)
                     if os.path.isdir(os.path.join(self.save_path, name))][0]))
            path = os.path.join(path, os.listdir(path)[0])
        genes_info = pd.read_csv(
          os.path.join(path, 'genes.tsv'), sep='\t', header=None)
        gene_names = genes_info.values[:, self.genecol].astype(np.str).ravel()
        if os.path.exists(os.path.join(path, 'barcodes.tsv')):
            self.barcodes = pd.read_csv(os.path.join(path, 'barcodes.tsv'),
                                        sep='\t', header=None)
        # print(genes_info)
        # self.gene_symbols = \
        #     genes_info.values[:, self.genecol].astype(np.str).ravel()
        expression_data = io.mmread(os.path.join(path, 'matrix.mtx')).T
        if self.dense:
            expression_data = expression_data.A
        else:
            expression_data = csr_matrix(expression_data)

        print("Finished preprocessing dataset")
        return expression_data, gene_names


class BrainSmallDataset(Dataset10X):
    r"""
    This dataset consists in 9,128 mouse brain cells profiled using
    `10x Genomics`_ is used as a complement of PBMC for our study of zero
    abundance and quality control metrics correlation with our generative
    posterior parameters. We derived quality control metrics using the
    cellrangerRkit R package (v.1.1.0). Quality metrics were extracted from
    CellRanger throughout the molecule specific information file. We kept the
    top 3000 genes by variance. We used the clusters provided by cellRanger
    for the correlation analysis of zero probabilities.

    Args:
        :save_path: Save path of raw data file. Default: ``'data/'``.

    Examples:
        >>> gene_dataset = BrainSmallDataset()

    .. _10x Genomics:
        https://support.10xgenomics.com/single-cell-gene-expression/datasets

    """

    def __init__(self, save_path='data/'):
        dataset = Dataset10X(filename="neuron_9k", save_path=save_path)
        self.save_path = save_path
        self.urls = ['https://github.com/YosefLab/scVI-data/raw/master/'
                     'brain_small_metadata.pickle']
        self.download_names = ['brain_small_metadata.pickle']
        self.download()

        metadata = pickle.load(open(os.path.join(
          self.save_path, 'brain_small_metadata.pickle'), 'rb'))
        labels = metadata['clusters'].loc[dataset.barcodes.values.ravel()] - 1

        self.raw_qc = metadata['raw_qc'].loc[dataset.barcodes.values.ravel()]
        self.qc_names = self.raw_qc.columns
        self.qc = self.raw_qc.values

        GeneExpressionDataset.__init__(
          self, dataset.Y, batch_indices=dataset.batch_indices, labels=labels)



import torch
from .gradient import gradient_ee


class ELoss(torch.autograd.Function):

    @staticmethod
    def forward(ctx, X, Lp, Wn, lam):
        ctx.X, ctx.Lp, ctx.Wn, ctx.lam = X, Lp, Wn, lam
        return torch.tensor(0)

    def backward(ctx, grad_output):
        X, Lp, Wn, lam = ctx.X, ctx.Lp, ctx.Wn, ctx.lam
        return gradient_ee(X, Lp, Wn, lam), None, None, None


def eloss(X, Lp, Wn, lam):
    return ELoss.apply(X, Lp, Wn, lam)



from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    "numpy>=1.16.3",
    "scipy>=1.1.0",
    "torch>=1.0.1",
    "matplotlib>=2.2.3",
    "tqdm>=4.26.0",
    "anndata>=0.6.19",
    "scikit-learn>=0.19.2",
    "h5py>=2.8.0",
    "pandas>=0.23.4",
    "loompy>=2.0.17",
    "jupyter>=1.0.0",
    "ipython>=6.5.0",
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author='Jiankang Xiong',
    author_email='hibearme@163.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    description="Scalable Visualization of Massive Single-Cell Data"
                "Using Neural Networks",
    install_requires=requirements,
    license="MIT license",
    long_description=readme,
    include_package_data=True,
    keywords='neuralee',
    name='neuralee',
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/HibearME/NeuralEE',
    version='0.1.2',
    zip_safe=False,
)



from .dataset import GeneExpressionDataset
import pandas as pd
import numpy as np
import os


class CsvDataset(GeneExpressionDataset):
    r""" Loads a `.csv` file.

    Args:
        :filename: Name of the `.csv` file.
        :save_path: Save path of the dataset. Default: ``'data/'``.
        :url: Url of the remote dataset. Default: ``None``.
        :new_n_genes: Number of subsampled genes. Default: ``600``.
        :subset_genes: List of genes for subsampling. Default: ``None``.
        :compression: For on-the-fly decompression of on-disk data.
                      If ‘infer’ and filepath_or_buffer
            is path-like, then detect compression from the following
            extensions: ‘.gz’, ‘.bz2’, ‘.zip’, or ‘.xz’
            (otherwise no decompression). If using ‘zip’, the ZIP file must
            contain only one data file to be read in.
            Default: ``None``.
        :batch_ids_file: Name of the `.csv` file with batch indices.
            File contains two columns. The first holds gene names and second
            holds batch indices - type int.
            The first row of the file is header.

    Examples:
        >>> # Loading a remote dataset
        >>> remote_url = "https://www.ncbi.nlm.nih.gov/geo/download/" \
        ... "?acc=GSE100866&format=file&file=" \
        ... "GSE100866%5FCBMC%5F8K%5F13AB%5F10X%2DRNA%5Fumi%2Ecsv%2Egz")
        >>> remote_csv_dataset = CsvDataset(
        ...     "GSE100866_CBMC_8K_13AB_10X-RNA_umi.csv.gz", save_path='data/',
        ...     compression='gzip', url=remote_url)
        >>> # Loading a local dataset
        >>> local_csv_dataset = CsvDataset(
        ...     "GSE100866_CBMC_8K_13AB_10X-RNA_umi.csv.gz",
        ...     save_path='data/', compression='gzip')

    """

    def __init__(self, filename, save_path='data/', url=None, new_n_genes=600,
                 subset_genes=None, compression=None, sep=',',
                 gene_by_cell=True, labels_file=None, batch_ids_file=None):
        self.download_name = filename  # The given csv file is
        self.save_path = save_path
        self.url = url
        self.compression = compression
        self.sep = sep
        # Whether the original dataset is genes by cells
        self.gene_by_cell = gene_by_cell
        self.labels_file = labels_file
        self.batch_ids_file = batch_ids_file

        data, gene_names, labels, cell_types, batch_ids = \
            self.download_and_preprocess()

        super().__init__(
            *GeneExpressionDataset.get_attributes_from_matrix(
                data, labels=labels,
                batch_indices=batch_ids if batch_ids is not None else 0),
            gene_names=gene_names, cell_types=cell_types)

        self.subsample_genes(new_n_genes, subset_genes)

    def preprocess(self):
        print("Preprocessing dataset")

        if self.gene_by_cell:
            data = pd.read_csv(
                os.path.join(self.save_path, self.download_name),
                sep=self.sep, index_col=0, compression=self.compression).T
        else:
            data = pd.read_csv(
                os.path.join(self.save_path, self.download_name),
                sep=self.sep, index_col=0, compression=self.compression)

        gene_names = np.array(data.columns, dtype=str)
        labels, cell_types, batch_ids = None, None, None
        if self.labels_file is not None:
            labels = pd.read_csv(
                os.path.join(self.save_path, self.labels_file), header=0,
                index_col=0)
            labels = labels.values
            cell_types = np.unique(labels)

        if self.batch_ids_file is not None:
            batch_ids = pd.read_csv(
                os.path.join(self.save_path, self.batch_ids_file), header=0,
                index_col=0)
            batch_ids = batch_ids.values

        data = data.values
        print("Finished preprocessing dataset")
        return data, gene_names, labels, cell_types, batch_ids


class BreastCancerDataset(CsvDataset):
    def __init__(self, save_path='data/'):
        super().__init__("Layer2_BC_count_matrix-1.tsv", save_path=save_path,
                         url="http://www.spatialtranscriptomicsresearch.org/"
                             "wp-content/uploads/2016/07/"
                             "Layer2_BC_count_matrix-1.tsv",
                         sep='\t', gene_by_cell=False)


class MouseOBDataset(CsvDataset):
    def __init__(self, save_path='data/'):
        super().__init__("Rep11_MOB_count_matrix-1.tsv", save_path=save_path,
                         url="http://www.spatialtranscriptomicsresearch.org/"
                             "wp-content/uploads/2016/07/"
                             "Rep11_MOB_count_matrix-1.tsv",
                         sep='\t', gene_by_cell=False)



import pickle
import os
import numpy as np
import pandas as pd

from .dataset import GeneExpressionDataset, arrange_categories
from .dataset10X import Dataset10X


class PbmcDataset(GeneExpressionDataset):
    r""" Loads pbmc dataset.

    We considered scRNA-seq data from two batches of peripheral blood
    mononuclear cells (PBMCs) from a healthy donor (4K PBMCs and 8K PBMCs).
    We derived quality control metrics using the cellrangerRkit
    R package (v. 1.1.0). Quality metrics were extracted from CellRanger
    throughout the molecule specific information file. After filtering, we
    extract 12,039 cells with 10,310 sampled genes and get biologically
    meaningful clusters with the software Seurat. We then filter genes that we
    could not match with the bulk data used for differential expression to be
    left with g = 3346.

    Args:
        :save_path: Save path of raw data file. Default: ``'data/'``.

    Examples:
        >>> gene_dataset = PbmcDataset()

    """

    def __init__(self, save_path='data/'):
        self.save_path = save_path
        self.urls = ['https://github.com/YosefLab/scVI-data/raw/master/'
                     'gene_info.csv',
                     'https://github.com/YosefLab/scVI-data/raw/master/'
                     'pbmc_metadata.pickle']

        self.download_names = ['gene_info_pbmc.csv', 'pbmc_metadata.pickle']
        self.download()
        self.de_metadata = pd.read_csv(
            os.path.join(self.save_path, 'gene_info_pbmc.csv'), sep=',')

        pbmc_metadata = pickle.load(
            open(os.path.join(self.save_path, 'pbmc_metadata.pickle'), 'rb'))

        pbmc = GeneExpressionDataset.concat_datasets(
            Dataset10X("pbmc8k", save_path=save_path),
            Dataset10X("pbmc4k", save_path=save_path)
        )
        self.barcodes = pd.concat(pbmc.barcodes).values.ravel().astype(str)
        super().__init__(pbmc.Y, pbmc.batch_indices, pbmc.labels,
                         pbmc.gene_names)

        dict_barcodes = dict(zip(self.barcodes, np.arange(len(self.barcodes))))
        subset_cells = []
        barcodes_metadata = \
            pbmc_metadata['barcodes'].index.values.ravel().astype(np.str)
        for barcode in barcodes_metadata:
            # barcodes with end -11 filtered on 10X website (49 cells)
            if barcode in dict_barcodes:
                subset_cells += [dict_barcodes[barcode]]
        self.update_cells(subset_cells=np.array(subset_cells))
        idx_metadata = \
            np.array([not barcode.endswith('11')
                      for barcode in barcodes_metadata], dtype=np.bool)
        self.design = pbmc_metadata['design'][idx_metadata]
        self.raw_qc = pbmc_metadata['raw_qc'][idx_metadata]
        self.qc_names = self.raw_qc.columns
        self.qc = self.raw_qc.values

        self.qc_pc = pbmc_metadata['qc_pc'][idx_metadata]
        self.normalized_qc = pbmc_metadata['normalized_qc'][idx_metadata]

        labels = pbmc_metadata['clusters'][idx_metadata] \
            .reshape(-1, 1)[:self.Y.shape[0]]
        self.labels, self.n_labels = arrange_categories(labels)
        self.cell_types = pbmc_metadata['list_clusters'][:self.n_labels]

        # only keep the genes for which we have de data
        genes_to_keep = list(self.de_metadata['ENSG'].values)
        # Non empty only for unit tests
        difference = list(set(genes_to_keep).difference(set(pbmc.gene_names)))
        for gene in difference:
            genes_to_keep.remove(gene)
        self.filter_genes(genes_to_keep)
        # this would only affect the unit tests
        self.de_metadata = self.de_metadata.head(len(genes_to_keep))


class PurifiedPBMCDataset(GeneExpressionDataset):
    r""" The purified PBMC dataset from: "Massively parallel digital
    transcriptional profiling of single cells".

    Args:
        :save_path: Save path of raw data file. Default: ``'data/'``.

    Examples:
        >>> gene_dataset = PurifiedPBMCDataset()

    """

    def __init__(self, save_path='data/', filter_cell_types=None):
        cell_types = np.array(["cd4_t_helper", "regulatory_t", "naive_t",
                               "memory_t", "cytotoxic_t", "naive_cytotoxic",
                               "b_cells", "cd4_t_helper", "cd34", "cd56_nk",
                               "cd14_monocytes"])
        # filter = np.arange(6) - for T cells:  np.arange(4) for T/CD4 cells
        if filter_cell_types:
            cell_types = cell_types[np.array(filter_cell_types)]

        datasets = []
        for cell_type in cell_types:
            dataset = Dataset10X(cell_type, save_path=save_path)
            dataset.cell_types = np.array([cell_type])
            datasets += [dataset]

        pbmc = GeneExpressionDataset.concat_datasets(
            *datasets, shared_batches=True)
        pbmc.subsample_genes(
            subset_genes=(np.array(pbmc.X.sum(axis=0)) > 0).ravel())
        super().__init__(pbmc.X, pbmc.batch_indices, pbmc.labels,
                         gene_names=pbmc.gene_names,
                         cell_types=pbmc.cell_types)



import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cm as cmx
import numpy as np


def scatter(X, labels=None, cell_types=None, title=None, s=1, fg_kwargs=dict(),
            size=1.0, lg_kwargs=dict()):
    if fg_kwargs == dict():
        fg_kwargs = {'dpi': 200}
    if lg_kwargs == dict():
        lg_kwargs = {'markerscale': 2, 'fontsize': 'xx-small'}

    N = X.shape[0]
    if labels is None:
        labels = np.zeros(N)
    N_sub = size if isinstance(size, int) else int(N * size)
    if N_sub != N:
        ind_sub = np.random.permutation(N)[: N_sub]
        X = X[ind_sub]
        labels = labels[ind_sub]
    label_unique = np.unique(labels)
    if cell_types is not None:
        assert len(cell_types) == len(label_unique), \
            'Labels do not correspond to cell types.'
    jet = plt.get_cmap('jet')  # 'nipy_spectral'
    cNorm = colors.Normalize(vmin=0, vmax=range(len(label_unique))[-1])
    scalarMap = cmx.ScalarMappable(norm=cNorm, cmap=jet)
    plt.figure(**fg_kwargs)
    for il, ll in enumerate(label_unique):
        XX = X[labels == ll, :]
        if cell_types is not None:
            plt.scatter(XX[:, 0], XX[:, 1], s=s, color=scalarMap.to_rgba(il),
                        label=cell_types[ll])
        else:
            plt.scatter(XX[:, 0], XX[:, 1], s=s, color=scalarMap.to_rgba(il),
                        label=str(ll))
    if len(label_unique) != 1:
        plt.legend(**lg_kwargs)
    plt.axis('off')
    if title is not None:
        plt.title(title)
    plt.tight_layout()


def scatter_without_outlier(X, n_outlier=1000, s=0.0001,
                            fg_kwargs={'figsize': [10, 10]},
                            labels=None, cell_types=None, lg_kwargs=dict()):
    median = np.median(X, axis=0)
    X_reduce = X - median
    d2 = (X_reduce ** 2).sum(axis=1)
    ind_remain = d2.argsort()[:: -1][n_outlier:]
    X_reduce = X_reduce[ind_remain]
    if labels is not None:
        labels = labels[ind_remain]
    scatter(X_reduce, s=s, fg_kwargs=fg_kwargs, labels=labels,
            cell_types=cell_types, lg_kwargs=lg_kwargs)


def scatter_with_colorbar(X, labels=None, cell_types=None, title=None, s=1,
                          fg_kwargs=dict(), size=1.0):
    if cell_types is not None:
        assert len(cell_types) == 2, \
            'It only supports polarizable cell types'
    if fg_kwargs == dict():
        fg_kwargs = {'dpi': 200}

    N = X.shape[0]
    N_sub = size if isinstance(size, int) else int(N * size)
    if N_sub != N:
        ind_sub = np.random.permutation(N)[: N_sub]
        X = X[ind_sub]
        labels = labels[ind_sub]
    plt.figure(**fg_kwargs)
    plt.scatter(X[:, 0], X[:, 1], s=s, c=labels)

    if cell_types is not None:
        cbar = plt.colorbar(orientation='horizontal', fraction=0.046, pad=0.04)
        maxl = labels.max()
        minl = labels.min()
        ticks = [minl + (maxl - minl) * 0.1, (minl + maxl) / 2,
                 minl + (maxl - minl) * 0.9]
        cbar.set_ticks(ticks)
        cbar.set_ticklabels([cell_types[0], 'Other', cell_types[1]])

    plt.axis('off')
    if title is not None:
        plt.title(title)
    plt.tight_layout()



import torch
import numpy as np
import math


def error_ee(X, Wp, Wn, lam):
    ker, sqd = ker_ee(X)
    error = Wp.view(-1).dot(sqd.view(-1)) + lam * Wn.view(-1).dot(ker.view(-1))
    return error, ker


def error_ee_from_ker(ker, sqd, Wp, Wn, lam):
    return Wp.view(-1).dot(sqd.view(-1)) + lam * Wn.view(-1).dot(ker.view(-1))


def ker_ee(X):
    sqd = sqdist(X)
    ker = torch.exp(-sqd)
    return ker, sqd


def sqdist(X):
    x = (X ** 2).sum(dim=1, keepdim=True)
    sqd = x - 2 * X @ X.t() + x.t()
    ind = torch.arange(X.shape[0]).tolist()
    sqd[ind, ind] = torch.zeros(
        X.shape[0], device=X.device, dtype=torch.float32)
    return sqd


def error_ee_cpu(X, Wp, Wn, lam):
    mem = 4
    N = X.shape[0]
    # This will fit in mem GB RAM
    B = math.floor((mem * 1024 ** 3) / (2 * N * 8))
    error = 0
    i1 = 0
    i2 = min(N, B)
    X2 = X ** 2
    x2 = X2.sum(axis=1, keepdims=True)

    while i1 < N:
        sqd = X2[i1: i2, :].sum(axis=1, keepdims=True) - \
            2 * X[i1: i2, :] @ X.T + x2.T
        ker = np.exp(-sqd)
        error += Wp[i1: i2, :].reshape(-1).dot(sqd.reshape(-1)) + \
            lam * Wn[i1: i2, :].reshape(-1).dot(ker.reshape(-1))
        i1 = i1 + B
        i2 = min(N, i1 + B)
    return error


def error_ee_cuda(X, Wp, Wn, lam):
    device = X.device
    mem = 2
    N = X.shape[0]
    # This will fit in mem GB RAM
    B = math.floor((mem * 1024 ** 3) / (2 * N * 8))
    error = 0
    i1 = 0
    i2 = min(N, B)
    X2 = X ** 2
    x2 = X2.sum(dim=1, keepdim=True)

    while i1 < N:
        sqd = X2[i1: i2, :].sum(dim=1, keepdim=True) - \
            2 * X[i1: i2, :] @ X.t() + x2.t()
        ker = (-sqd).exp()
        error += Wp[i1: i2, :].to(device).view(-1).dot(sqd.view(-1)) + \
            lam * Wn[i1: i2, :].to(device).view(-1).dot(ker.view(-1))
        i1 = i1 + B
        i2 = min(N, i1 + B)
    return error



import numpy as np
from scipy.sparse import csr_matrix


def nnsqdist(X):

    N = X.shape[0]
    x2 = (X ** 2).sum(axis=1, keepdims=True)
    sqd = x2 - 2 * X @ X.T + x2.T
    np.fill_diagonal(sqd, 0)
    nn = sqd.argsort(axis=1)[:, 1: N]
    D2 = np.take_along_axis(sqd, nn, axis=1)

    return D2, nn, sqd


def ea(X, K):
    """Gaussian entropic affinities. This operation is in 'cpu'.

    This computes the matrix W of Gaussian entropic affinities (EAs) and the
    corresponding bandwidth values s for a dataset X and a desired perplexity
    K. Reference from:
    https://eng.ucmerced.edu/people/vladymyrov/papers/icml13.pdf

    """

    D2, nn, sqd = nnsqdist(X)
    N, k = D2.shape
    b = np.zeros(N, dtype=np.float32)
    Wp = np.zeros((N, k), dtype=np.float32)
    logK = np.log(K).astype(np.float32)
    B, D2 = _eabounds(logK, D2)  # Log-beta bounds
    # Point order: distance to Kth nn
    p = np.argsort(D2[:, int((np.ceil(K))) - 1])
    p = p.tolist()
    j = p[0]
    b0 = B[j, :].mean()
    p.append(-1)  # Initialization
    for i in range(N):
        # Compute log-beta & EAs for each point
        b[j], Wp[j, :] = _eabeta(D2[j, :], b0, logK, B[j, :])
        b0 = b[j]
        j = p[i + 1]  # Next point
    W = csr_matrix((Wp.reshape(-1),
                    nn.reshape(-1),
                    np.arange(N * k + 1, step=k)), shape=(N, N)).toarray()
    # s = 1 / np.sqrt((2 * np.exp(b))) # Bandwidths from log-beta values
    return W, sqd


def _eabounds(logK, D2):

    eps = 1.1921e-7  # 2 ** (-23)
    N = D2.shape[1]
    logN = np.log(N).astype(np.float32)
    logNK = logN - logK

    delta2 = D2[:, 1] - D2[:, 0]
    # Ensure delta2 >= eps
    ind = np.nonzero(delta2 < eps)[0]
    i = 2
    flag = 1
    while ind.shape[0] != 0:
        if (i + 1) > np.exp(logK) and flag:
            D2[ind, 0] = D2[ind, 0] * 0.99
            flag = 0
        delta2[ind] = D2[ind, i] - D2[ind, 0]
        ind = np.nonzero(delta2 < eps)[0]
        i = i + 1

    deltaN = D2[:, N - 1] - D2[:, 0]

    # Compute p1(N,logK)
    if logK > np.log(np.sqrt(2 * N)):
        p1 = 3 / 4
    else:
        p1 = 1 / 4
        for _ in range(100):
            e = -p1 * np.log(p1 / N) - logK
            g = -np.log(p1 / N) + 1
            p1 = p1 - e / g
        p1 = 1 - p1 / 2

    bU1 = (2 * np.log(p1 * (N - 1) / (1 - p1))) / delta2
    bL1 = (2 * logNK / (1 - 1 / N)) / deltaN
    bL2 = np.sqrt(2 * np.sqrt(logNK)) / (D2[:, N - 1] ** 2 - D2[:, 0] ** 2)
    B = np.log(np.stack((np.maximum(bL1, bL2), bU1), axis=1))

    return B, D2


def _eabeta(d2, b0, logK, B):
    """
    Computes the values of beta and the corresponding Gaussian affinities for
    one point. It does root finding with Newton's method embedded in a
    bisection loop to ensure global convergence.

    """""
    eps = 1.1921e-7  # 2 ** (-23)
    realmin = np.float32(1e-45)
    maxit = 20  # Max. no. iterations before a bisection
    tol = 1e-3  # Convergence tolerance to stop iterating

    if (b0 < B[0] or b0 > B[1]):
        b = (B[0] + B[1]) / np.float32(2)
    else:
        b = b0

    i = 1  # maxit counter
    # Inside the loop, pbm is a flag to detect any numerical problems (zero
    # gradient, infinite function value, etc.).
    while True:
        bE = np.exp(b)
        pbm = 0

        # Compute the function value: m0, m1v, m1 are O(N)
        ed2 = np.exp(-d2 * bE)
        m0 = ed2.sum()
        if m0 < realmin:  # Numerical error
            m0 = realmin
            e = -logK
            pbm = 1
        else:
            m1v = ed2 * d2 / m0
            m1 = m1v.sum()
            e = bE * m1 + np.log(m0) - logK
        if abs(e) < tol:
            break

        # Very narrow bounds, no need to iterate.
        # This can happen if K is very small.
        if B[1] - B[0] < 10 * eps:
            break

        # Update the bounds
        if (e < 0 and b <= B[1]):
            B[1] = b
        elif (e > 0 and b >= B[0]):
            B[0] = b

        pbm = pbm or np.isinf(e) or e < -logK or e > np.log(d2.shape[0]) - logK

        if not pbm:
            if i == maxit:  # Exceeded maxit, bisection step
                b = (B[0] + B[1]) / np.float32(2)
                i = 1
                continue
            # Compute the gradient: m2 is O(N)
            eg2 = bE ** np.float32(2)
            m2 = (m1v * d2).sum()
            m12 = m1 ** np.float32(2) - m2
            g = eg2 * m12
            if g == 0:
                pbm = 1

        if pbm:
            b = (B[0] + B[1]) / np.float32(2)
            i = 1
            continue

        # Newton step ok, update bounds
        p = -e / g
        b = b + p

        if (b < B[0] or b > B[1]):  # Out of bounds, bisection step
            b = (B[0] + B[1]) / np.float32(2)
            i = 0
        i = i + 1

    W = ed2 / m0  # Affinities
    return b, W


def x2p(X, perplexity=20.0):
    tol = 1e-3
    n = X.shape[0]
    P = np.zeros((n, n), dtype=np.float32)
    beta = np.ones(n, dtype=np.float32)
    logU = np.log(perplexity).astype(np.float32)

    x2 = (X ** 2).sum(axis=1, keepdims=True)
    D = x2 - 2 * X @ X.T + x2.T
    for i in range(n):
        betamin = -np.inf
        betamax = np.inf

        Di = np.delete(D[i], i)
        H, thisP = Hbeta(Di, beta[i])

        Hdiff = H - logU
        tries = 0
        while abs(Hdiff) > tol and tries < 50:

            if Hdiff > 0:
                betamin = beta[i]
                if np.isinf(betamax):
                    beta[i] = beta[i] * 2
                else:
                    beta[i] = (beta[i] + betamax) / 2
            else:
                betamax = beta[i]
                if np.isinf(betamin):
                    beta[i] = beta[i] / 2
                else:
                    beta[i] = (beta[i] + betamin) / 2

            H, thisP = Hbeta(Di, beta[i])
            Hdiff = H - logU
            tries = tries + 1

        P[i] = np.insert(thisP, i, 0)
        if i < n - 1:
            beta[i + 1] = beta[i]
    return P, D


def Hbeta(D, beta):
    # P = np.exp(-D * beta)
    # sumP = P.sum()
    # H = np.log(sumP) + beta * (D * P).sum() / sumP
    # P = P / sumP
    realmax = np.float32(3.4028e+38)
    shift = np.log(realmax) / np.float32(2) - (-D * beta).max()
    ff = np.exp(-D * beta + shift)
    zz = ff.sum()
    P = ff / zz
    H = np.log(zz) - shift + beta * (D * P).sum()
    return H, P



import h5py
import numpy as np
from scipy.sparse import csr_matrix, vstack
from sklearn.preprocessing import StandardScaler
import os

from .dataset import GeneExpressionDataset

batch_idx_10x = [1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
                 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
                 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1,
                 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
                 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
                 1, 1, 1, 1, 1, 1, 1]


class BrainLargeDataset(GeneExpressionDataset):
    r""" Loads brain-large dataset.

    This dataset contains 1.3 million brain cells from `10x Genomics`_.
    We randomly shuffle the data to get a 1M subset of cells and order genes
    by variance to retain first 10,000 and then 720 sampled variable genes.
    This dataset is then sampled multiple times in cells for the runtime and
    goodness-of-fit analysis. We report imputation scores on the 10k cells
    and 720 genes samples only.

    Args:
        :save_path: Save path of raw data file. Default: ``'data/'``.

    Examples:
        >>> gene_dataset = BrainLargeDataset()

    .. _10x Genomics:
        https://support.10xgenomics.com/single-cell-gene-expression/datasets

    """

    def __init__(self, subsample_size=None, save_path='data/',
                 nb_genes_kept=720, max_cells=None):
        self.max_cells = max_cells
        self.subsample_size = subsample_size
        self.save_path = save_path
        self.nb_genes_kept = nb_genes_kept
        self.url = "http://cf.10xgenomics.com/samples/cell-exp/1.3.0/" \
                   "1M_neurons/1M_neurons_filtered_gene_bc_matrices_h5.h5"
        # originally: "1M_neurons_filtered_gene_bc_matrices_h5.h5"

        self.download_name = "genomics.h5"
        Xs = self.download_and_preprocess()
        batch_indices = np.zeros((Xs.shape[0], 1))
        labels = np.zeros((Xs.shape[0], 1))
        super().__init__(Xs, batch_indices=batch_indices, labels=labels)

    def preprocess(self):
        print("Preprocessing Brain Large data")

        filtered_matrix_h5 = os.path.join(self.save_path, self.download_name)
        with h5py.File(filtered_matrix_h5) as f:
            dset = f["mm10"]
            n_genes, n_cells = f["mm10"]["shape"]
            if self.subsample_size is None:
                self.subsample_size = n_cells
            indptr = dset['indptr'][...]

            ns_cells = min(10000, n_cells)  # TODO : remove
            ns_indptr = indptr[:(ns_cells + 1)]
            ns_nnz = ns_indptr[-1]
            ns_data = dset["data"][:ns_nnz].astype(np.float32)
            ns_indices = dset["data"][:ns_nnz]
            ns_sparse = csr_matrix((ns_data, ns_indices, ns_indptr),
                                   shape=(ns_cells, n_genes))
            ns_dense = np.log(1 + ns_sparse.toarray())

            # Use standard scaler to order select genes by variance
            std_scaler = StandardScaler(with_mean=False)
            std_scaler.fit(ns_dense)
            subset_genes = \
                np.argsort(std_scaler.var_)[::-1][:self.nb_genes_kept]

            nb_matrices = []
            nb_cells = 100000
            nb_iters = int(self.subsample_size / nb_cells) + \
                (self.subsample_size % nb_cells > 0)
            for i in range(nb_iters):
                nb_indptr = indptr[(i * nb_cells):((1 + i) * nb_cells + 1)]
                nb_nnz_a = nb_indptr[0]
                nb_nnz_b = nb_indptr[-1]
                nb_indptr = (nb_indptr - nb_nnz_a).astype(np.int32)
                nb2_cells = len(nb_indptr) - 1
                nb_data = dset["data"][nb_nnz_a:nb_nnz_b].astype(np.float32)
                nb_indices = \
                    dset["indices"][nb_nnz_a:nb_nnz_b].astype(np.int32)
                nb_sparse = csr_matrix(
                    (nb_data, nb_indices, nb_indptr),
                    shape=(nb2_cells, n_genes))
                nb_filtered = nb_sparse[:, subset_genes]
                del nb_sparse
                nb_matrices.append(nb_filtered)
                print("loaded {} / {} cells".format(i * nb_cells + nb2_cells,
                                                    self.subsample_size))
                if self.max_cells and \
                   i * nb_cells + nb2_cells >= self.max_cells:
                    break

        matrix = vstack(nb_matrices)
        good_cells = matrix.sum(axis=1) > 0
        good_cells = np.squeeze(np.asarray(good_cells))
        print("excluding {} cells with zero genes expressed".format(
            len(good_cells) - good_cells.sum()))
        matrix = matrix[good_cells, :]

        print("%d cells subsampled" % matrix.shape[0])
        print("%d genes subsampled" % matrix.shape[1])

        return matrix
